/**
 * Copyright (c) 2009-2010 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.URISyntaxException;
import java.net.UnknownHostException;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 *         This class manages the test harness.
 * 
 */
public class Director extends Thread {

    public static final String VERSION = "2010-04-05.1";

    public static final String NAME = Director.class.getName();

    protected static final String SIMPLE_NAME = Director.class
            .getSimpleName();

    protected static SimpleLogger logger = SimpleLogger.getSimpleLogger();

    protected Configuration config;

    protected Controller[] controllers;

    protected Timer timer;

    protected File resultsDirectory;

    public Director(String[] args) throws IOException, URISyntaxException {
        // base configuration - we change this later on
        config = new Configuration();
        config.setLogger(logger);
        initConfiguration(args);

        resultsDirectory = new File("cpox-" + VERSION + "/"
                + Utilities.formatDateTime(config.getStart()));
        resultsDirectory.mkdirs();

        logger.info("director hostname = "
                + InetAddress.getLocalHost().getHostName());
        logger.info(getVersionMessage());
    }

    /**
     * @param args
     * @throws URISyntaxException
     * @throws IOException
     * @throws FileNotFoundException
     * 
     */
    private void initConfiguration(String[] args)
            throws URISyntaxException, FileNotFoundException, IOException {
        // use system properties as a basis
        // this allows any number of properties at the command-line,
        // using -DPROPNAME=foo
        config.load(System.getProperties());

        // all args are assumed to be property files,
        // so add them to the base configuration
        try {
            // must pass properties to the new instance
            config = new Configuration(config);
        } catch (Exception e) {
            throw new FatalException(e);
        }

        // iterate through input files, adding each one
        File file;
        for (int i = 0; i < args.length; i++) {
            file = new File(args[i]);
            if (!file.exists()) {
                logger.warning("skipping " + args[i]
                        + ": file does not exist.");
                continue;
            }
            if (!file.canRead()) {
                logger.warning("skipping " + args[i]
                        + ": file cannot be read.");
                continue;
            }
            if (file.isDirectory()) {
                logger.warning("skipping " + args[i]
                        + ": is a directory.");
                continue;
            }
            logger.info("processing: " + args[i]);
            config.load(new FileInputStream(file));
        }

        // make the configuration final
        config.configure();
    }

    public static void main(String[] args) throws Exception {
        System.err.println(getVersionMessage());

        Director rl = new Director(args);
        rl.run();
        rl.join();
    }

    protected static String getVersionMessage() {
        return getVersionMessage(SIMPLE_NAME);
    }

    protected static String getVersionMessage(String _name) {
        return _name + " starting, version " + VERSION + " on "
                + System.getProperty("java.version") + " ("
                + System.getProperty("java.runtime.name") + ")" + " "
                + System.getProperty("file.encoding");
    }

    public void run() {
        try {
            recordConfiguration();
        } catch (IOException e) {
            throw new FatalException(e);
        }

        timer = new Timer();

        startControllers();

        // monitor until all controllers have stopped
        long lastReported = System.currentTimeMillis();
        while (pollControllers()) {
            // report ongoing performance stats
            if (System.currentTimeMillis() - lastReported > 20 * 1000) {
                logger.info(""
                        + (controllers.length * config
                                .getWorkersPerController())
                        + " threads at "
                        + String.format("%.0f", timer
                                .getEventsPerSecond())
                        + " requests/sec, with " + timer.getErrorCount()
                        + " errors");
                lastReported = System.currentTimeMillis();
            }

            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                // ignore
            }
        }

        logger.fine("exiting");
        timer.stop();

        pollControllers();

        try {
            recordResults();
        } catch (IOException e) {
            logger.logException(e);
        }

        // TODO: tell controllers to remove local results?

    }

    /**
     * @throws IOException
     * 
     */
    private void recordConfiguration() throws IOException {
        PrintStream out = newLogFile("configuration.properties");
        config.list(out);
        out.flush();
        out.close();
    }

    /**
     * @param _name
     * @return
     * @throws IOException
     * @throws FileNotFoundException
     */
    private PrintStream newLogFile(String _name) throws IOException,
            FileNotFoundException {
        File log = new File(resultsDirectory, _name);
        if (!log.createNewFile()) {
            throw new IOException("error creating file: "
                    + log.getCanonicalPath());
        }
        logger.info(log.getCanonicalPath());
        PrintStream out = new PrintStream(log);
        return out;
    }

    /**
     * @return
     */
    private boolean pollControllers() {
        boolean active = false;
        long requests;
        requests = 0;
        for (int i = 0; i < controllers.length; i++) {
            if (controllers[i].isAlive() && controllers[i].isActive()) {
                active = true;
            }
            requests += controllers[i].getRequestCount();
        }
        timer.setEventCount(requests);
        return active;
    }

    /**
     * @throws IOException
     * 
     */
    private void recordResults() throws IOException {
        PrintStream out = newLogFile("results.csv");
        Timer threadTimer = gatherControllerResults(out);
        out.flush();
        out.close();

        checkResults(threadTimer);

        // record configuration and results in summary document
        out = newLogFile("summary.txt");
        tee(out, String.format("count: %21d", timer.getEventCount()));
        tee(out, String.format("errors: %20d", timer.getErrorCount()));
        tee(out, String.format("duration (ms): %13.0f", timer
                .getDurationMilliseconds()));
        tee(out, String.format("throughput (req/sec): %6.0f", timer
                .getEventsPerSecond()));
        // use the worker thread timer for average response time
        tee(out, String.format("average (ms): %14.0f", threadTimer
                .getMeanOverallMilliseconds()));
        out.flush();
        out.close();
    }

    /**
     * @param out
     * @param _message
     */
    private void tee(PrintStream out, String _message) {
        logger.info(_message);
        out.println(_message);
    }

    /**
     * @param out
     * @return
     * @throws IOException
     */
    private Timer gatherControllerResults(PrintStream out)
            throws IOException {
        // retrieve and record final results from controllers
        BufferedReader br;
        String workerName;
        String line;
        int workers = config.getWorkersPerController();
        boolean firstLine = false;
        // accumulate basic statistics in a timer
        String fields[];
        Timer timer = new Timer(0, 0);
        for (int i = 0; i < controllers.length; i++) {
            for (int j = 0; j < workers; j++) {
                workerName = controllers[i].getWorkerName(j);
                br = controllers[i].getResultsReader(j);
                // skip first line (titles) after i==0
                if ((0 == i && 0 == j)) {
                    firstLine = true;
                } else {
                    br.readLine();
                }
                // add worker name and append to output
                while ((line = br.readLine()) != null) {
                    if (firstLine) {
                        out.println("worker" + "," + line);
                        firstLine = false;
                    } else {
                        out.println(workerName + "," + line);
                        fields = line.split(",");
                        if ("-1".equals(fields[4])) {
                            // error
                            timer.incrementErrorCount();
                        } else {
                            timer.incrementEventCount();
                        }
                        timer.incrementDuration(Long.valueOf(fields[5]));
                    }
                }
            }
        }
        return timer;
    }

    /**
     * @param threadTimer
     */
    protected void checkResults(Timer threadTimer) {
        // sanity checks
        if (timer.getEventCount() != threadTimer.getEventCount()) {
            logger.warning("event count mismatch: "
                    + timer.getEventCount() + " != "
                    + threadTimer.getEventCount() + " (workers)");
        }
        if (timer.getErrorCount() != threadTimer.getErrorCount()) {
            logger.warning("error count mismatch: "
                    + timer.getErrorCount() + " != "
                    + threadTimer.getErrorCount() + " (workers)");
        }
    }

    /**
     * 
     */
    private void startControllers() {
        // trivial case: one local controller
        // TODO: handle network-connected controllers
        int size = 1;
        controllers = new Controller[size];

        if (controllers.length < 1) {
            throw new FatalException("no controllers available");
        }
        for (int i = 0; i < controllers.length; i++) {
            try {
                controllers[i] = new Controller(this, i);
            } catch (UnknownHostException e) {
                throw new FatalException(e);
            }
            controllers[i].start();
            logger.info("controller " + i + ": "
                    + controllers[i].getHostname());
        }

        logger.fine("waiting for controllers");

        for (int i = 0; i < controllers.length; i++) {
            while (!(controllers[i].isAlive() && controllers[i]
                    .isActive())) {
                try {
                    Thread.sleep(10);
                } catch (InterruptedException e) {
                    // ignore
                }
            }
        }

    }

    /**
     * @return
     */
    public Configuration getConfig() {
        return config;
    }

    /**
     * @param e
     */
    public void reportException(Exception e) {
        timer.incrementErrorCount();
        // TODO: include reporter's host name and/or controller index?
        if (e instanceof ActionException
                && ((ActionException) e).isTerse()) {
            logger.warning(e.getMessage());
        } else {
            logger.logException(e);
        }
        if (e instanceof FatalException) {
            shutdown();
            return;
        }
    }

    /**
     * 
     */
    private void shutdown() {
        if (null != controllers) {
            for (int i = 0; i < controllers.length; i++) {
                controllers[i].shutdown();
            }
        }
        Runtime.getRuntime().exit(0);
    }
}
