/**
 * Copyright (c) 2009 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import info.bliki.wiki.model.WikiModel;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedList;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public class WikiToZip {

    protected static final String SIMPLE_NAME = WikiToZip.class
            .getSimpleName();

    /**
     * 
     */
    private static final String TITLE = "title";

    /**
     * 
     */
    private static final String PAGE = "page";

    // number of zip entries overflows at 2^16 = 65536
    // ref: http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4828461
    // (supposed to be fixed, but isn't)
    static final int MAX_ENTRIES = 65536 - 2;

    private static final String ROOT = "mediawiki";

    private static final String NS_XML = "http://www.w3.org/XML/1998/namespace";

    private static LinkedList<String> path;

    private static StringBuilder article;

    private static ZipOutputStream output;

    private static int entries;

    private static String title;

    private static XmlPullParser xpp;

    private static int outputCount = 0;

    static SimpleLogger logger = SimpleLogger.getSimpleLogger();

    private static int errors = 0;

    private static int pages = 0;

    private static String namespace;

    private static String language;

    private static XmlPullParserFactory factory;

    private static XmlPullParser parser;

    /**
     * @param args
     * @throws Exception
     */
    public static void main(String[] args) throws Exception {
        System.err.println(Director.getVersionMessage(SIMPLE_NAME));

        // expect input wikimedia XML on stdin
        Properties properties = new Properties();
        factory = XmlPullParserFactory.newInstance(properties
                .getProperty(XmlPullParserFactory.PROPERTY_NAME), null);
        factory.setNamespaceAware(true);
        xpp = factory.newPullParser();
        xpp.setInput(new InputStreamReader(System.in));

        // TODO feature isn't supported by xpp3 - look at xpp5?
        // xpp.setFeature(XmlPullParser.FEATURE_DETECT_ENCODING, true);
        // TODO feature isn't supported by xpp3 - look at xpp5?
        // xpp.setFeature(XmlPullParser.FEATURE_PROCESS_DOCDECL, true);
        xpp.setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES, true);

        logger.configureLogger(new Properties());
        logger.info(Director.getVersionMessage(SIMPLE_NAME));

        process();
        logger.info("finished " + pages + " pages with " + errors
                + " errors");
    }

    /**
     * @throws IOException
     * @throws XmlPullParserException
     */
    private static void process() throws XmlPullParserException,
            IOException {
        // transform to final output
        int event;
        path = new LinkedList<String>();
        article = null;
        output = null;
        title = null;

        logger.info("starting loop");

        while (true) {
            event = xpp.next();
            switch (event) {
            case XmlPullParser.END_DOCUMENT:
                processEndDocument();
                // exit the loop
                return;
            case XmlPullParser.END_TAG:
                processEndElement(xpp.getName());
                break;
            case XmlPullParser.START_TAG:
                processStartElement(xpp.getName());
                break;
            case XmlPullParser.TEXT:
                if (null != article) {
                    String name = path.getLast();
                    if ("comment".equals(name) || "text".equals(name)) {
                        // parse comment elements
                        // parse text elements
                        article.append(parse(xpp.getText()));
                    } else {
                        article
                                .append(Utilities
                                        .escapeXml(xpp.getText()));
                    }
                }
                break;
            default:
                throw new IOException("unexpected event: " + event
                        + " at " + xpp.getPositionDescription());
            }
        }
    }

    /**
     * @param text
     * @return
     * @throws IOException
     */
    private static String parse(String text) throws IOException {
        if (null == text || "".equals(text.trim())) {
            return null;
        }
        // parse wiki markup to xml
        // TODO: this is slow with bliki - might need concurrency
        // use a new object every time, to prevent leaks
        // no doubt this makes it slower...
        String xml = new WikiModel("${image}", "${title}").render(text);

        if (null == xml || "".equals(xml.trim())) {
            return xml;
        }

        // verify xml is well-formed
        try {
            // use this xpp object to check output from the wikimedia parser
            parser = factory.newPullParser();
            parser
                    .setInput(new StringReader("<dummy>" + xml
                            + "</dummy>"));
            parser.setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES,
                    true);
            int event;
            String temp;
            char[] chars;
            int c;
            while (true) {
                // with some Japanese text, next() throws
                // ArrayIndexOutOfBoundsException
                try {
                    event = parser.next();
                } catch (ArrayIndexOutOfBoundsException e) {
                    throw new XmlPullParserException(e.getMessage(),
                            parser, null);
                }
                switch (event) {
                case XmlPullParser.END_DOCUMENT:
                    // exit the loop
                    return xml;
                case XmlPullParser.END_TAG:
                    parser.getName();
                    parser.getNamespace();
                    parser.getText();
                    break;
                case XmlPullParser.START_TAG:
                    parser.getName();
                    parser.getNamespace();
                    parser.getText();
                    break;
                case XmlPullParser.TEXT:
                    temp = parser.getText();
                    if (null != temp) {
                        chars = temp.toCharArray();
                        // xpp3 doesn't check codepoint values
                        // check them to avoid XDMP errors
                        for (int i = 0; i < chars.length; i++) {
                            c = chars[i];
                            // #x9 | #xA | #xD
                            // | [#x20-#xD7FF]
                            // | [#xE000-#xFFFD]
                            // | [#x10000-#x10FFFF]
                            // this implementation is abbreviated
                            if (9 == c || 10 == c || 13 == c || c > 31) {
                                continue;
                            }
                            throw new XmlPullParserException(
                                    "bad codepoint value: " + c, parser,
                                    null);
                        }
                    }
                    break;
                default:
                    throw new IOException("unexpected event: " + event
                            + " at " + parser.getPositionDescription());
                }
            }
        } catch (XmlPullParserException e) {
            logger.warning(title + ": " + e.getMessage());
            errors++;
            return Utilities.escapeXml(text);
        }
    }

    /**
     * @param name
     * @throws IOException
     */
    private static void processEndElement(String name) throws IOException {
        // logger.info(name);
        if (!path.getLast().equals(name)) {
            throw new IOException("found " + name + " expected "
                    + path.getLast() + "; " + title + "; " + article);
        }
        path.removeLast();

        if (null == article) {
            return;
        }

        article.append(xpp.getText());

        // look for end of article
        if (!PAGE.equals(name)) {
            return;
        }

        // check zip archive count
        if (null == output || entries >= MAX_ENTRIES) {
            nextOutput();
        }

        boolean encodeTitle = false;
        URI uri = null;
        if (encodeTitle) {
            // try encoding the entry name
            try {
                // this form of URI() does escaping nicely
                uri = new URI(null, title, null);
            } catch (URISyntaxException e) {
                try {
                    // URI(schema, ssp, fragment) constructor cannot handle
                    // ssp = 2008-11-07T12:23:47.617766-08:00/1
                    // (despite what the javadoc says)...
                    // in this situation, treat the path as the fragment.
                    uri = new URI(null, null, title);
                } catch (URISyntaxException e1) {
                    throw new IOException(e);
                }
            }
        }

        // write article to zip archive
        // include the language in the title
        ZipEntry entry = new ZipEntry(language + "wiki/"
                + (encodeTitle ? uri.toString() : title));
        entry.setCompressedSize(-1);
        output.putNextEntry(entry);
        output.write(article.toString().getBytes());
        output.closeEntry();
        output.flush();
        entries++;

        // ready for the next page
        article = null;
        entry = null;
    }

    /**
     * @throws IOException
     * 
     */
    private static void nextOutput() throws IOException {
        if (null != output) {
            output.flush();
            output.close();
        }
        // expect about 300 zip archives - lots of margin
        String path = String.format("%s-%08x%s", WikiToZip.class
                .getSimpleName(), outputCount, ".zip");
        logger.info(path + " " + (null == title ? "" : title));
        File outFile = new File(path);
        output = new ZipOutputStream(new FileOutputStream(outFile));
        output.setLevel(9);
        outputCount++;
        entries = 0;
    }

    /**
     * @param name
     * @throws IOException
     * @throws XmlPullParserException
     */
    private static void processStartElement(String name)
            throws IOException, XmlPullParserException {
        // logger.info(name);
        path.add(name);
        // look for start of article
        if (ROOT.equals(name)) {
            namespace = xpp.getNamespace();
            language = xpp.getAttributeValue(NS_XML, "lang");
            return;
        }

        if (PAGE.equals(name)) {
            if (null != article) {
                throw new IOException("article not null at start of page");
            }
            // this is clumsy, but should work ok
            article = new StringBuilder("<"
                    + PAGE
                    // propagate the XML namespace
                    + (null == namespace ? ""
                            : (" xmlns=\"" + namespace + "\""))
                    // propagate the xml:lang attribute
                    + (null == language ? ""
                            : (" xml:lang=\"" + language + "\""))
                    // end of the start tag
                    + ">");
            pages++;
            return;
        }

        if (null != article && !xpp.isEmptyElementTag()) {
            // write empty elements via end-element, only.
            // note that attributes are still ok in this case
            article.append(xpp.getText());
        }

        if (!TITLE.equals(name)) {
            return;
        }

        // create zip entry when we see the title element
        title = xpp.nextText().trim();
        article.append(Utilities.escapeXml(title));
        // this puts us at the end element for title
        processEndElement(name);
    }

    /**
     * @throws IOException
     */
    private static void processEndDocument() throws IOException {
        if (0 != path.size()) {
            throw new IOException("document end before end tag ("
                    + path.size() + ") " + path.getLast() + " "
                    + xpp.getPositionDescription());
        }
        if (null != article) {
            throw new IOException("article not null at end of document: "
                    + title + "; " + article.toString() + "; "
                    + xpp.getPositionDescription());
        }
        if (null != output) {
            output.flush();
            output.close();
        }
    }

}
